"""Collection of utilities of the freva command line argument parser."""

from __future__ import annotations
import abc
import argparse
from copy import copy
import logging
import os
from pathlib import Path
import sysconfig
from typing import Optional, Type

import lazy_import
from evaluation_system.misc import logger

freva = lazy_import.lazy_module("freva")
config = lazy_import.lazy_module("evaluation_system.misc.config")


def get_cli_class(name: str) -> Optional[Type[BaseParser]]:
    """Get core module or cli extension.

    Parameters
    ----------
    name: str
        name of the cli sub module that is to be imported. This can be
        either a cli sub module for the freva core (databrowser, plugin etc)
        or an extension. Extensions are assumed to follow the following nameing
        conventions: :py:mod:`<name>.app`. The ``app`` sub module must contain
        a class named :py:class:`Cli`.

    Returns
    -------
    module: The imported module
    """
    mod_name = name.replace("-", "_")
    try:  # First try importing a module extensions
        mod = __import__(f"{mod_name}.cli", fromlist=[""])
    except ImportError:
        try:  # Freva core module?
            mod = __import__(f"freva.cli.{mod_name}", fromlist=[""])
        except ImportError:
            return None
    if hasattr(mod, "Cli") and hasattr(mod.Cli, "desc"):
        return mod.Cli
    return None


class BaseParser(metaclass=abc.ABCMeta):
    """Base class that is used to construct a valid freva cli tool.

    This class can be used to extent the freva commands with additional sub
    commands.

    Parameters
    ----------
    parser: argparse.ArgumentParser, default: None
        Already parsed arguments if, None given (default) a new
        :py:class:`ArgumentParser` will should be constructed.

    Attributes
    ----------
    desc: str
        The class should have the :py:attribute:`desc` attribute set. This
        attribute is used to display a short summary over what this command
        is supposed to do.


    Example
    -------
    Suppose your tool named ``mytool`` should be made a freva cli extension the
    then the ``mytool`` library should define a command line interface that
    resides in the ``mytool.cli`` sub module of ``mytool``. Additionally your
    ``mytool.cli`` sub module has to define a class called ``Cli``. The content
    of the ``Cli`` class could like the following:

    .. code-block:: python

        import argparse
        from pathlib import Path
        import sys
        from typing import Optional

        from ferva.cli import BaseParser

        class Cli(BaseParser):

            desc = "Apply mytool."

            def __init__(parser: Optional[argparse.ArgumentParser] = None):

                super().__init__("mytool", parser)
                self.parser.add_argument("path",
                                         type=Path,
                                         help="The first argument"
                )

            def do_something(path: Path) -> list[str]:
                '''Do the actual work, for example list a directory.'''
                return sorted(path.iterdir())

        def main():
            '''The main method calling the cli from a stand alone tool.'''

            arg_p = Cli()
            args = arg_p.parse_args(sys.argv[1:])
            arg_p.do_something(args.path)

    Following this outline all that would be needed to do is creating an entry
    point named ``freva-mytool`` pointing to ``mytool.cli:main``. This would
    be sufficent to creagte a freva mytool sub command.
    """

    desc: str = ""
    """The short describtion of a freva command."""

    def __init__(
        self,
        parser: Optional[argparse.ArgumentParser] = None,
        command: str = "freva",
    ):
        self.parser = parser or argparse.ArgumentParser(
            prog=command,
            description=self.desc,
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        )

    @property
    def logger(self) -> logging.Logger:
        """Use evaluation_system logger in all classes using ths class."""
        return logger

    def set_debug(self, debug: bool) -> None:
        """Set the logger level to DEBUG."""
        if debug is True:
            self.logger.setLevel(logging.DEBUG)

    def parse_args(self, argv: Optional[list[str]] = None) -> argparse.Namespace:
        """Parse the command line arguments."""
        args, other_args = self.parser.parse_known_args(argv)
        self.kwargs = {k: v for (k, v) in args._get_kwargs() if k != "apply_func"}
        self.kwargs["other_args"] = other_args
        self.set_debug(self.kwargs.pop("debug", False))
        return args


class SubCommandParser(BaseParser):
    """Base class for all freva sub command line argument parsers."""

    def __init__(
        self,
        parser: Optional[argparse.ArgumentParser] = None,
        sub_parsers: Optional[dict[str, Type[BaseParser]]] = None,
        command: str = "freva",
    ) -> None:
        """Create the sub-command parsers."""
        super().__init__(parser, command)
        self.subparsers = self.parser.add_subparsers(help="Available sub-commands:")
        self.sub_commands = sub_parsers or self.get_subcommand_parsers()
        self.help = self.get_subcommand_help(self.sub_commands)
        for cmd, subparser in self.sub_commands.items():
            self.parse_subcommand(cmd, subparser)

    def parse_subcommand(self, command: str, cli_class: Type[BaseParser]) -> None:
        """Parse the databrowser command."""

        call_parsers = self.subparsers.add_parser(
            command,
            description=cli_class.desc,
            help=cli_class.desc,
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        )
        cli_class(call_parsers)

    def _usage(self, *args: Optional[str], **kwargs: Optional[str]) -> None:
        """Exit with usage message."""
        self.parser.error(
            "the following sub-commands are "
            f"required: {', '.join(self.sub_commands.keys())}"
        )

    @staticmethod
    def get_subcommand_parsers() -> dict[str, Type[BaseParser]]:
        """Create the help strings of the available sub commands."""
        path_list = [
            Path(p)
            for p in (os.environ.get("PATH") or os.defpath).split(os.pathsep)
            + [sysconfig.get_paths().get("scripts", "")]
            if p and Path(p).is_dir()
        ]
        parsers = {}
        for _dir in path_list:
            for file in _dir.iterdir():
                if file.name.startswith("freva-"):
                    _, _, sub_cmd = file.name.partition("-")
                    if sub_cmd not in parsers:
                        parser_class = get_cli_class(sub_cmd)
                        if parser_class is not None:
                            parsers[sub_cmd] = parser_class
        return parsers

    @classmethod
    def get_subcommand_help(
        cls,
        parsers: Optional[dict[str, Type[BaseParser]]] = None,
    ) -> dict[str, str]:
        """Create the help strings of the available sub commands."""
        parsers = parsers or cls.get_subcommand_parsers()
        return {h: mod.desc for (h, mod) in parsers.items() if hasattr(mod, "desc")}


class BaseCompleter:
    """Base class for command line argument completers."""

    def __init__(
        self,
        metavar: str,
        argv: list[str],
        choices: Optional[dict[str, tuple[str, str]]] = None,
        shell: str = "bash",
        strip: bool = False,
        flags_only: bool = False,
    ):
        self.choices = choices or {}
        self.strip = strip
        self.metavar = metavar
        self.argv = argv
        self.flags_only = flags_only
        if shell == "zsh":
            self.get_print = self._print_zsh
        elif shell == "fish":
            self.get_print = self._print_fish
        else:
            self.get_print = self._print_default

    def _print_zsh(self, choices: dict[str, tuple[str, str]]) -> list[str]:
        out = []
        for key, (_help, func) in choices.items():
            if self.metavar != "databrowser" or key.startswith("-"):
                out.append(f"{key}[{_help}]{func}")
            else:
                out.append(f"{key}: {_help}")
        return out

    def _print_fish(self, choices: dict[str, tuple[str, str]]) -> list[str]:
        out = []
        for key, (_help, func) in choices.items():
            out.append(f"{key}: {_help}")
        return out

    def _print_default(self, choices: dict[str, tuple[str, str]]) -> list[str]:
        out = []
        for key, (_help, _) in choices.items():
            if self.metavar == "databrowser" and not key.startswith("-"):
                out.append(f"{key}: {_help}")
            else:
                out.append(key)
        return out

    def _get_databrowser_choices(self) -> dict[str, tuple[str, str]]:
        """Get the choices for databrowser command."""

        facet_args = []
        for arg in self.argv:
            if len(arg.split("=")) == 2:
                facet_args.append(arg)
        facets = BaseCompleter.arg_to_dict(facet_args)
        search = freva.databrowser(attributes=False, all_facets=True, **facets)
        choices = {}
        for att, values in search.items():
            if att not in facets:
                keys = ",".join([v for n, v in enumerate(values)])
                choices[att] = (keys, "")
        return choices

    def _get_plugin_choices(self) -> dict[str, tuple[str, str]]:
        """Get the choices for the plugin command."""

        docs: dict[str, dict[str, str]] = {}
        desc: dict[str, dict[str, str]] = {}
        plugins: dict[str, str] = {}
        try:
            plugin_cache = freva.read_plugin_cache()
        except Exception:
            plugin_cache = []
        for plugin in plugin_cache:
            plugins[plugin.name] = plugin.description
            docs[plugin.name] = {}
            desc[plugin.name] = {}
            for param, doc, value in plugin.parameters:
                desc[plugin.name][param] = doc
                docs[plugin.name][param] = value
        args = [arg for arg in self.argv if not arg.startswith("-") and arg != "plugin"]
        choices = {}
        if not args:
            # No plugin name was given
            return {pl: (help, "") for pl, help in plugins.items()}
        try:
            plugin_config = docs[args[0]]
            setup = desc[args[0]]
        except KeyError:
            # Wrong plugin name
            return {}
        options = []
        for key in plugin_config.keys():
            option_present = False
            for arg in args:
                if arg.startswith(f"{key}="):
                    option_present = True
                    break
            if not option_present:
                options.append(f"{key}=")
        for option in options:
            opt = option.strip("=")
            if "file" in setup[opt].lower() or "dir" in setup[opt].lower():
                choices[opt] = (setup[opt], ":file:_files")
            else:
                choices[opt] = (setup[opt], "")
        return choices

    @property
    def command_choices(self) -> dict[str, tuple[str, str]]:
        """Get the command line arguments for all sub commands."""

        choices = {}
        if self.flags_only:
            return self.choices
        if self.metavar == "databrowser":
            choices = self._get_databrowser_choices()
        elif self.metavar == "plugin":
            choices = self._get_plugin_choices()
        return {**self.choices, **choices}

    def formated_print(self) -> None:
        """Print all choices to be processed by the shell completion function."""

        out = self.get_print(self.command_choices)
        for line in out:
            if line.startswith("-") and self.strip and not self.flags_only:
                continue
            elif not line.startswith("-") and self.flags_only:
                continue
            print(line)

    @staticmethod
    def arg_to_dict(
        args: Optional[list[str]], append: bool = False
    ) -> dict[str, list[str]]:
        """Convert a parsed arguments with key=value pairs to a dictionary.

        Parameters:
        ----------
        args:
            Collection of arguments that are converted to dict, the arguments
            should be of form key=value
        append:
            If a key occurs twice convert first value to list and append second
            value to list, the default behaviour is updating the entry with
            the second value
        Returns:
        --------
        dict: Dictionariy representation of key=value pairs
        """
        out_dict: dict[str, list[str]] = {}
        for arg in args or []:
            key, _, value = arg.partition("=")
            if append and key in out_dict:
                out_dict[key].append(value)
            else:
                out_dict[key] = [value]
        return out_dict

    @classmethod
    def _get_choices_from_parser(
        cls,
        parser: argparse.ArgumentParser,
        argv: list[str],
    ) -> dict[str, tuple[str, str]]:
        choices: dict[str, tuple[str, str]] = {}
        for action in parser._actions:
            action_type = ""
            if isinstance(action, argparse._SubParsersAction):
                if argv:
                    cmd = argv.pop(0)
                    sub_parser = action.choices.get(cmd, "")
                    if sub_parser:
                        choices.update(cls._get_choices_from_parser(sub_parser, argv))
                else:
                    for ch, _parser in action.choices.items():
                        choices[ch] = _parser.description or "", action_type
            if action.help == argparse.SUPPRESS:
                # This is an option that is not exposed to users
                continue
            if action.type == Path:
                action_type = ":file:_files"
            if action.option_strings:
                choice = action.option_strings[0]
            else:
                choice = action.dest
            if choice.lower() not in (
                "facet",
                "facets",
                "tool-name",
                "options",
                "==suppress==",
                "-h",
            ):
                choices[choice] = (action.help or "", action_type)
        return choices

    @classmethod
    def get_args_of_subcommand(cls, argv: list[str]) -> dict[str, tuple[str, str]]:
        """Get all possible arguments from a freva sub-command."""
        sub_command = argv.pop(0)
        parser_class = get_cli_class(sub_command)
        if parser_class is None:
            return {}
        cli_parser = parser_class()
        return cls._get_choices_from_parser(cli_parser.parser, argv)

    @classmethod
    def parse_choices(cls, argv: list[str]) -> BaseCompleter:
        """Create the completion choices from given cmd arguments."""
        parser = argparse.ArgumentParser(
            description="Get choices for command line arguments"
        )
        parser.add_argument(
            "command", help="First command, freva, freva-histor etc", type=str
        )
        parser.add_argument("--shell", help="Shell in use", default="bash", type=str)
        parser.add_argument(
            "--strip",
            help="Do not print options starting with -",
            default=False,
            action="store_true",
        )
        parser.add_argument(
            "--flags-only",
            help="Only print options starting with -",
            default=False,
            action="store_true",
        )
        app, args = parser.parse_known_args(argv)
        main_choices = {
            k: (v, "") for k, v in SubCommandParser.get_subcommand_help().items()
        }
        if app.command == "freva" and not args:
            return cls(app.command, [], main_choices, shell=app.shell, strip=app.strip)
        if app.command != "freva":
            sub_cmd = "-".join(app.command.split("-")[1:])
            args.insert(0, sub_cmd)
        choices = {}
        if args[0] in main_choices:
            choices = cls.get_args_of_subcommand(copy(args))
        choices = {k: v for (k, v) in choices.items() if k not in args}
        return cls(
            args[0],
            args,
            choices,
            shell=app.shell,
            strip=app.strip,
            flags_only=app.flags_only,
        )


def print_choices(arguments: list[str]) -> None:
    """Print completion choices based of command line arguments.

    Parameters:
    -----------
    arguments:
        List of command line arguments, that are evaluated by the choice printer
    """
    argv = [arg.strip() for arg in arguments if arg.strip()]
    comp = BaseCompleter.parse_choices(argv)
    comp.formated_print()



"""Update user data in the apache solr data search server."""

from __future__ import annotations
from dataclasses import dataclass
import logging
from pathlib import Path
import os
import shutil
from typing import Callable, Optional, Union
import warnings

import lazy_import
from evaluation_system.misc import logger
from evaluation_system.misc.exceptions import ConfigurationException, ValidationError


User = lazy_import.lazy_class("evaluation_system.model.user.User")
config = lazy_import.lazy_module("evaluation_system.misc.config")
SolrCore = lazy_import.lazy_class("evaluation_system.model.solr_core.SolrCore")
DataReader = lazy_import.lazy_class("evaluation_system.api.user_data.DataReader")
get_output_directory = lazy_import.lazy_function(
    "evaluation_system.api.user_data.get_output_directory"
)

__all__ = ["UserData"]


@dataclass
class UserData:
    """Data class that handles user data requests. With help of this class
    users can add their own data to the databrowser, (re)-index data in the
    databrowser or delete data in the databrowser."""

    @property
    def user_dir(self) -> Path:
        """Get the user data directory for the user."""
        try:
            return get_output_directory() / f"user-{User().getName()}"
        except ConfigurationException:
            config.reloadConfiguration()
            return get_output_directory() / f"user-{User().getName()}"

    def _validate_user_dirs(self, *crawl_dirs: os.PathLike) -> tuple[Path, ...]:
        root_path = self.user_dir
        user_paths: tuple[Path, ...] = ()
        for crawl_dir in crawl_dirs or (root_path,):
            crawl_dir = Path(crawl_dir or root_path).expanduser().absolute()
            try:
                _ = crawl_dir.relative_to(root_path)
            except ValueError as error:
                raise ValidationError(
                    f"You are only allowed to crawl data in {root_path}"
                ) from error
            user_paths += (crawl_dir,)
        return user_paths

    @staticmethod
    def _set_add_method(how: str) -> Callable[[os.PathLike, os.PathLike], None]:
        choices = "copy, link, move, symlink, cp, ln, mv"
        if how in ["copy", "cp"]:
            return shutil.copy
        if how in ["symlink", "ln"]:
            return os.symlink
        if how in ["move", "mv"]:
            return shutil.move
        if how in ["link"]:
            return os.link
        raise ValueError(f"Invalid Method: valid methods are {choices}")

    def add(
        self,
        product: str,
        *paths: os.PathLike,
        how: str = "copy",
        override: bool = False,
        **defaults: str,
    ) -> None:
        """Add custom user files to the databrowser.

        To be able to add data to the databrowser the file names must
        follow a strict standard and the files must reside in a
        specific location. This ``add`` method takes care about the correct
        file naming and location. No pre requirements other than the file has
        to be a valid ``netCDF`` or ``grib`` file are assumed. In other words
        this method places the user data with the correct naming structure to
        the correct location.

        Parameters
        ----------
        product: str
            Product search key the newly added data can be found.
        *paths: os.PathLike
            Filename(s) or Directories that are going to be added to the
            databrowser. The files will be added into the central user
            directory and named according the CMOR standard. This ensures
            that the data can be added into the databrowser.
            **Note:** Once the data has been added into the databrowser it can
            be found via the ``user-<username>`` project.
        how: str, default: copy
            Method of how the data is added into the central freva user directory.
            Default is copy, which means your data files will be replicated.
            To avoid a this redundancy you can set the ``how`` keyword to
            ``symlink`` for symbolic links or ``link`` for creating hard links
            to create symbolic links or ``move`` to move the data into the central
            user directory entirely.
        override: bool, default: False
            Replace existing files in the user data structre
        experiment: str, default: None
            By default the method tries to deduce the *experiment* information from
            the metadata. To overwrite this information the *experiment* keyword
            should be set.
        institute: str, default: None
            By default the method tries to deduce the *institute* information from
            the metadata. To overwrite this information the *institute* keyword
            should be set.
        model: str, default: None
            By default the method tries to deduce the *model* information from
            the metadata. To overwrite this information the *model* keyword
            should be set.
        variable: str, default: None
            By default the method tries to deduce the *variable* information from
            the metadata. To overwrite this information the *variable* keyword
            should be set.
        time_frequency: str, default: None
            By default the method tries to deduce the *time_frequency* information
            from the metadata. To overwrite this information the *time_frequency*
            keyword should be set.
        ensemble: str, default: None
            By default the method tries to deduce the *ensemble* information from
            the metadata. To overwrite this information the *ensemble* keyword
            should be set.

        Raises
        ------
        ValueError: If metadata is insufficient, or product key is empty


        Example
        -------

        Suppose you've gotten data from somewhere and want to add this data into the
        databrowser to make it accessible to others. In this specific
        example we assume that you have stored your `original` data in the
        ``/tmp/my_awesome_data`` folder.
        E.g ``/tmp/my_awesome_data/outfile_0.nc...tmp/my_awesome_data/outfile_9.nc``
        The routine will try to gather all necessary metadata from the files. You'll
        have to provide additional metadata if mandatory keywords are missing.
        To make the routine work you'll have to provide the ``institute``, ``model``
        and ``experiment`` keywords:

        .. execute_code::

            from freva import UserData, databrowser
            user_data = UserData()
            # You can also provide wild cards to search for data
            user_data.add("eur-11b", "/tmp/my_awesome_data/outfile_?.nc",
                              institute="clex", model="UM-RA2T",
                              experiment="Bias-correct")
            # Check the databrowser if the data has been added
            for file in databrowser(experiment="bias*"):
                print(file)

        By default the data is copied. By using the ``how`` keyword you can
        also link or move the data.
        """
        crawl_dirs: list[Path] = []
        facets = (
            "experiment",
            "institute",
            "model",
            "variable",
            "time_frequency",
            "ensemble",
        )
        search_keys = {k: defaults[k] for k in facets if defaults.get(k)}
        search_keys["product"] = product
        search_keys["project"] = f"user-{User().getName()}"
        search_keys["realm"] = "user_data"
        search_keys.setdefault("ensemble", "r0i0p0")
        for path in paths:
            p_path = Path(path).expanduser().absolute()
            u_reader = DataReader(p_path, **search_keys)
            for file in u_reader:
                new_file = u_reader.file_name_from_metdata(file, override=override)
                new_file.parent.mkdir(exist_ok=True, parents=True, mode=0o2775)
                if new_file.exists() and override:
                    new_file.unlink()
                self._set_add_method(how)(file, new_file)
                if new_file.parent not in crawl_dirs:
                    crawl_dirs.append(new_file.parent)
        if not crawl_dirs:
            warnings.warn("No files found", category=UserWarning)
            return
        self.index(*crawl_dirs)

    def delete(self, *paths: os.PathLike, delete_from_fs: bool = False) -> None:
        """Delete data from the databrowser.

        The methods deletes user data from the databrowser.

        Parameters
        ----------

        *paths: os.PathLike
            Filename(s) or Directories that are going to be from  the
            databrowser.
        delete_from_fs: bool, default : False
            Do not only delete the files from the databrowser but also from their
            central location where they have been added to.

        Raises
        ------
        ValidationError:
            If crawl_dirs do not belong to current user.

        Example
        -------

        Any data in the central user directory that belongs to the user can
        be deleted from the databrowser and also from the central data location:

        .. execute_code::

            from freva import UserData
            user_data = UserData()
            user_data.delete(user_data.user_dir)

        """
        solr_core = SolrCore(core="files")
        for path in paths:
            for file in DataReader(Path(path).expanduser().absolute()):
                self._validate_user_dirs(file)
                solr_core.delete_entries(str(file))
                if delete_from_fs:
                    file.unlink()

    def index(
        self,
        *crawl_dirs: os.PathLike,
        dtype: str = "fs",
        continue_on_errors: bool = False,
    ) -> None:
        """Index and add user output data to the databrowser.

        This method can be used to update the databrowser for existing user data

        Parameters
        ----------
        crawl_dirs:
            The data path(s) that needs to be crawled.
        dtype:
            The data type, currently only files on the file system are supported.
        continue_on_errors:
            Continue indexing on error.

        Raises
        ------
        ValidationError:
            If crawl_dirs do not belong to current user.

        Example
        -------

        If data has been removed from the databrowser it can be re added using
        the ``index`` method:

        .. execute_code::

            from freva import UserData
            user_data = UserData()
            user_data.index()

        """
        if dtype not in ("fs",):
            raise NotImplementedError("Only data on POSIX file system is supported")
        log_level = logger.level
        try:
            logger.setLevel(logging.ERROR)
            print("Status: crawling ...", end="", flush=True)
            solr_core = SolrCore(core="latest")
            for crawl_dir in self._validate_user_dirs(*crawl_dirs):
                data_reader = DataReader(crawl_dir)
                solr_core.load_fs(
                    crawl_dir,
                    chunk_size=1000,
                    abort_on_errors=not continue_on_errors,
                    drs_type=data_reader.drs_specification,
                )
            print("ok", flush=True)
        finally:
            logger.setLevel(log_level)



from __future__ import annotations
import argparse
import sys
from typing import Any, Optional

import lazy_import
from evaluation_system import __version__
from evaluation_system.misc import logger
from .utils import BaseParser, BaseCompleter

freva = lazy_import.lazy_module("freva")


class Cli(BaseParser):
    """Class that constructs the Databrowser Argument Parser."""

    desc = "Find data in the system."

    def __init__(
        self,
        parser: Optional[argparse.ArgumentParser] = None,
    ):
        """Construct the databrwoser sub arg. parser."""
        super().__init__(parser, "freva-databrowser")
        self.parser.add_argument(
            "--multiversion",
            default=False,
            action="store_true",
            help="Select not only the latest version.",
        )
        self.parser.add_argument(
            "--relevant-only",
            default=False,
            action="store_true",
            help="Show only search with results >1 possible values",
        )
        self.parser.add_argument(
            "--batch-size",
            default=5000,
            type=int,
            help="Number of files to retrieve",
        )
        self.parser.add_argument(
            "--count",
            default=False,
            action="store_true",
            help="Show the number of files for each search result",
        )
        self.parser.add_argument(
            "--attributes",
            default=False,
            action="store_true",
            help=(
                "Retrieve all possible attributes for the current "
                "search instead of the files. Deprecated, to be removed in "
                "v2304.0.0"
            ),
        )
        self.parser.add_argument(
            "--all-facets",
            default=False,
            action="store_true",
            help=(
                "retrieve all facets (attributes & values) instead of "
                "the files (deprecated, to be removed in v2304.0.0.0)"
            ),
        )
        self.parser.add_argument(
            "--facet",
            default=None,
            type=str,
            action="append",
            help=("Retrieve values of given facet instead of files"),
        )
        self.parser.add_argument(
            "--facet-limit",
            type=int,
            help="Limit the number of output facets.",
            default=sys.maxsize,
        )
        self.parser.add_argument(
            "--time-select",
            type=str,
            help="Operator that specifies how the time period is selected.",
            choices=["flexible", "strict", "file"],
            default="flexible",
        )
        self.parser.add_argument(
            "--debug",
            "-v",
            "-d",
            "--verbose",
            help="Use verbose output.",
            action="store_true",
            default=False,
        )
        self.parser.add_argument(
            "facets",
            nargs="*",
            help="Search facet(s)",
            type=str,
            metavar="facets",
        )
        self.parser.set_defaults(apply_func=self.run_cmd)

    @staticmethod
    def run_cmd(
        args: argparse.Namespace,
        other_args: Optional[list[str]] = None,
        **kwargs: Optional[Any],
    ) -> None:
        """Call the databrowser command and print the results."""
        facets: dict[str, Any] = BaseCompleter.arg_to_dict(args.facets, append=True)
        if args.relevant_only:
            logger.warning(
                "The relevant-only flag will be made deprecated",
            )
        if args.attributes:
            logger.warning(
                "The attributes flag will be made deprecated",
            )
        facet_limit = kwargs.pop("facet_limit")
        for key in (
            "facets",
            "all_facets",
            "count",
            "attributes",
            "relevant_only",
            "batch_size",
        ):
            _ = kwargs.pop(key, "")
        for key, values in facets.items():
            if len(values) == 1:
                facets[key] = values[0]
        merged_args: dict[str, Any] = {**kwargs, **facets}
        if args.all_facets:
            logger.warning(
                "The all-facets flag will be made deprecated use"
                "user --facet '*' or --facet all instead"
            )
            args.facet = "*"
        if args.count:
            _ = merged_args.pop("facet", "")
            out = freva.count_values(facet=args.facet, **merged_args)
        elif args.facet:
            _ = merged_args.pop("facet", "")
            out = freva.facet_search(facet=args.facet, **merged_args)
        elif args.attributes:
            out = freva.databrowser(attributes=True, **merged_args)
        else:
            out = freva.databrowser(batch_size=args.batch_size, **merged_args)
        # flush stderr in case we have something pending
        sys.stderr.flush()
        if isinstance(out, dict):
            # We have facet values as return values
            for att, values in out.items():
                facet_limit = facet_limit or len(values) + 1
                try:
                    keys = ",".join(
                        [
                            f"{k} ({c})"
                            for n, (k, c) in enumerate(values.items())
                            if n < facet_limit
                        ]
                    )
                except AttributeError:
                    keys = ",".join(
                        [v for n, v in enumerate(values) if n < facet_limit]
                    )
                if facet_limit < len(values):
                    keys += ",..."
                print(f"{att}: {keys}", flush=True)
            return
        if args.attributes:
            print(", ".join(out), flush=True)
            return
        if args.count:
            print(out, flush=True)
        else:
            for key in out:
                print(str(key), flush=True)


def main(argv: Optional[list[str]] = None) -> None:
    """Wrapper for entry point script."""
    cli = Cli()
    cli.parser.add_argument(
        "-V",
        "--version",
        action="version",
        version="%(prog)s {version}".format(version=__version__),
    )
    args = cli.parse_args(argv or sys.argv[1:])
    try:
        cli.run_cmd(args, **cli.kwargs)
    except KeyboardInterrupt:  # pragma: no cover
        print("KeyboardInterrupt, exiting", file=sys.stderr, flush=True)
        sys.exit(130)



import tempfile
import shutil
from pathlib import Path
from subprocess import run, PIPE
import os
import time

from evaluation_system.api.plugin import PluginAbstract
from evaluation_system.api.parameters import (
    ParameterDictionary,
    Integer,
    Float,
    String,
    InputDirectory,
    Directory,
)

from evaluation_system.model.user import User
from evaluation_system.model.db import UserDB


class DummyPluginFolders(PluginAbstract):
    """Stub class for implementing the abstract one"""

    __short_description__ = "A dummy plugin with outputdir folder"
    __long_description__ = ""
    __version__ = (0, 0, 0)
    __tags__ = ["foo"]
    __category__ = "statistical"
    __name__ = "DummyPluginFolders"
    __parameters__ = ParameterDictionary(
        String(name="variable", default="tas", help="An input variable"),
        Directory(
            name="outputdir",
            default="$USER_OUTPUT_DIR/$SYSTEM_DATETIME",
            mandatory=False,
            help="The default output directory",
        ),
    )
    _runs = []
    _template = "${number} - $something - $other"
    tool_developer = {"name": "DummyUser", "email": "data@dkrz.de"}

    def run_tool(self, config_dict=None):
        DummyPluginFolders._runs.append(config_dict)
        tool_path = Path(__file__).parent / "plugin_env" / "bin" / "python"
        res = run(["which", "python"], stdout=PIPE, stderr=PIPE)
        assert "plugin_env" in os.environ["PATH"]
        print(f"DummyPluginFolders tool was run with: {config_dict}")
        return {
            "/tmp/dummyfile1": dict(type="plot"),
            "/tmp/dummyfile2": dict(type="data"),
        }



"""
Created on 11.03.2013

@author: Sebastian Illing / estani

This package encapsulate access to a solr instance
"""

from __future__ import annotations
import urllib
from typing import cast, Union, List, NamedTuple
from typing_extensions import Literal
import warnings

from evaluation_system.model.solr_core import SolrCore
from evaluation_system.misc import logger, utils

SolrResponse = NamedTuple(
    "SolrResponse",
    [
        ("num_objects", int),
        ("start", int),
        ("exact", bool),
        ("docs", List[str]),
    ],
)


warnings.warn(
    "The evaluation_system.model.solr module will be removed from v2304.0.0",
    category=PendingDeprecationWarning,
)


class SolrFindFiles(object):
    """Encapsulate access to Solr like the find files command"""

    def __init__(self, core=None, host=None, port=None, get_status=False):
        """Create the connection pointing to the proper solr url and core.
        The default values of these parameters are setup in evaluation_system.model.solr_core.SolrCore
        and read from the configuration file.

        :param core: name of the solr core that will be used.
        :param host: hostname of the machine where the solr core is to be found.
        :param port: port number of the machine where the solr core is to be found.
        :param get_status: if the core should be contacted in an attempt to get more metadata.
        """
        self.solr = SolrCore(core, host=host, port=port, get_status=get_status)

    def __str__(self):  # pragma: no cover
        return "<SolrFindFiles %s>" % self.solr

    def _to_solr_query(self, partial_dict: dict[str, Union[str, list[str]]]) -> str:
        """Creates a Solr query assuming the default operator is "AND". See schema.xml for that."""
        params = []
        partial_dict = self._add_time_query(partial_dict)
        # these are special Solr keys that we might get and we assume are not meant for the search
        special_keys = ("q", "fl", "fq", "facet.limit", "sort")
        logger.debug(partial_dict)
        for key, value in partial_dict.items():
            if key in special_keys:
                params.append((key, value))
            else:
                if key.endswith("_not_"):
                    # handle negation
                    key = "-" + key[:-5]
                if isinstance(value, list):
                    # implies an or
                    constraint = " OR ".join(["%s:%s" % (key, v) for v in value])
                else:
                    constraint = "%s:%s" % (key, value)
                params.append(
                    (
                        "fq",
                        constraint,
                    )
                )
        logger.debug(params)
        return urllib.parse.urlencode(params)

    def _get_file_query_parameters(
        self,
        uniq_key: Literal["file", "uri"] = "file",
        **search_dict: Union[str, list[str]],
    ) -> str:
        partial_dict = search_dict.copy()
        for key in ("start", "row"):
            _ = partial_dict.pop("start", None)
        for key, value in {
            "q": "*:*",
            "fl": f"{uniq_key}",
            "sort": f"{uniq_key} desc",
        }.items():
            partial_dict.setdefault(key, value)
        if "text" in partial_dict:
            partial_dict["q"] = partial_dict.pop("text")
        return self._to_solr_query(partial_dict)

    def _retrieve_metadata(
        self, uniq_key: Literal["file", "uri"] = "file", **search_dict: str
    ) -> SolrResponse:
        """Retrieve metadata from databrowser.

        Parameters
        ----------


        uniq_key: str, default: file
            The unique key the solr server should be queried for.
        **search_dict: str
            Search query parameter

        Returns
        -------
        evaluation_system.model.solr.SolrResponse:
          NamedTuple of metadata on the search query results.
        """
        query = self._get_file_query_parameters(uniq_key=uniq_key, **search_dict)
        anw = self.solr.get_json("select?facet=true&rows=0&%s" % query)["response"]
        return SolrResponse(
            num_objects=anw["numFound"],
            start=anw["start"],
            exact=anw["numFoundExact"],
            docs=anw["docs"],
        )

    def _search(
        self,
        batch_size=10000,
        latest_version=False,
        uniq_key="file",
        rows=None,
        **partial_dict,
    ):
        """This encapsulates the Solr call to get documents and returns an iterator providing the. The special
        parameter _retrieve_metadata will affect the first value returned by the iterator.

        :param batch_size: the amount of files to be buffered from Solr.
        :param latest_version: if the search should *try* to find the latest version from all contained here. Please note
         that we don't use this anymore. Instead we have 2 cores and this is defined directly in :class:`SolrFindFiles.search`.
         It was changed because it was slow and required too much memory.
        known beforehand how many values are going to be returned, even before getting them all. To avoid this we might
        implement a result set object. But that would break the find_files compatibility.
        """
        offset = int(partial_dict.pop("start", "0"))
        query = self._get_file_query_parameters(uniq_key=uniq_key, **partial_dict)
        metadata = self._retrieve_metadata(uniq_key=uniq_key, **partial_dict)
        if rows:
            results_to_visit = min(metadata.num_objects, rows)
        else:
            results_to_visit = metadata.num_objects
        while results_to_visit > 0:
            batch_size = min(batch_size, results_to_visit)
            answer = self.solr.get_json(
                "select?start=%s&rows=%s&%s" % (offset, batch_size, query)
            )
            offset = answer["response"]["start"]
            iter_answer = answer["response"]["docs"]
            for item in iter_answer:
                yield item[uniq_key]
                results_to_visit -= 1
            offset += batch_size

    @staticmethod
    def _add_time_query(
        search_dict: dict[str, Union[str, list[str]]]
    ) -> dict[str, Union[list[str], str]]:
        """Add a potential time query string to the search dict."""
        time_subset = cast(str, search_dict.pop("time", ""))
        operator = cast(str, search_dict.pop("time_select", ""))
        if time_subset:
            start, _, end = time_subset.lower().partition("to")
            start = utils.convert_str_to_timestamp(start.strip() or "0", "")
            end = utils.convert_str_to_timestamp(end.strip() or start, "")
            if not start or not end:
                raise ValueError("Invalid time string")
            time = f"{{!field f=time op={operator}}}[{start} TO {end}]"
            search_dict["fq"] = time
        return search_dict

    @classmethod
    def get_metadata(
        cls,
        uniq_key: Literal["file", "uri"] = "file",
        latest_version: bool = True,
        **search_dict: str,
    ) -> SolrResponse:
        """Retrieve metadata from databrowser.

        Parameters
        ----------

        **search_dict: str
            Search query parameter

        Returns
        -------
        evaluation_system.model.solr.SolrResponse:
          NamedTuple of metadata on the search query results.
        """
        if latest_version:
            solrcore = cls(core="latest")
        else:
            solrcore = cls(core="files")
        return solrcore._retrieve_metadata(uniq_key=uniq_key, **search_dict)

    @staticmethod
    def search(latest_version=True, **partial_dict):
        """It mimics the same :class:`evaluation_system.model.file.DRSFile.search` behavior.
        The implementation contacts the required Solr cores instead of contacting the file system.

        :param latest_version: defines if looking for the latest version of a file only, or for any.
        :param partial_dict: the search dictionary for solr. It might also contain some special values as
         defined in :class:`SolrFindFiles._search`
        :returns: An iterator over the results."""
        # use defaults, if other required use _search in the SolrFindFiles instance
        if latest_version:
            s = SolrFindFiles(core="latest")
        else:
            s = SolrFindFiles(core="files")
        return s._search(**partial_dict)

    def _facets(self, latest_version=False, facets=None, **partial_dict):
        if facets and not isinstance(facets, list):
            if "," in facets:
                # we assume multiple values here
                facets = [f.strip() for f in facets.split(",")]
            else:
                facets = [facets]

        if "text" in partial_dict:
            partial_dict.update({"q": partial_dict.pop("text")})
        else:
            partial_dict.update({"q": "*:*"})

        query = self._to_solr_query(partial_dict)

        if facets is None:
            # get all minus what we don't want
            facets = self.solr.get_solr_fields() - set(
                [
                    "",
                    "_version_",
                    "file_no_version",
                    "level",
                    "timestamp",
                    "time",
                    "creation_time",
                    "source",
                    "version",
                    "uri",
                    "file",
                    "file_name",
                ]
            )

        if facets:
            query += (
                "&facet=true&facet.sort=index&facet.mincount=1&facet.field="
                + "&facet.field=".join(facets)
            )

        if latest_version:  # pragma: no cover (see above)
            query += "&group=true&group.field=file_no_version&group.facet=true"

        answer = self.solr.get_json("select?facet=true&rows=0&%s" % query)
        # TODO: why is there a language facit in the solr serach?
        answer = answer["facet_counts"]["facet_fields"]
        try:
            del answer["language"]
        except KeyError:
            pass
        return answer

    @staticmethod
    def facets(latest_version=True, facets=None, facet_limit=-1, **partial_dict):
        # use defaults, if other required use _search in the SolrFindFiles instance
        if latest_version:
            s = SolrFindFiles(core="latest")
        else:
            s = SolrFindFiles(core="files")
        return s._facets(facets=facets, latest_version=False, **partial_dict)



"""
Created on 18.05.2016

@author: Sebastian Illing
"""
from functools import partial
import os
import mock
import multiprocessing as mp
from pathlib import Path
import pytest
import time
from subprocess import Popen
from evaluation_system.tests import run_cli, similar_string
from evaluation_system.misc.exceptions import (
    PluginNotFoundError,
    ValidationError,
)


@mock.patch("os.getpid", lambda: 12345)
def test_cli(dummy_plugin, capsys, dummy_config, caplog):
    from freva.cli.plugin import main as plugin_cli
    from evaluation_system.misc import config

    with pytest.raises(ValidationError):
        plugin_cli(["dummyplugin"])
    plugin_cli(["dummyplugin", "the_number=13"])
    output = capsys.readouterr().out
    assert "the_number" in output
    assert "13" in output
    assert os.getpid() == 12345
    out_path = (
        Path(dummy_config.get("scheduler_output_dir"))
        / "dummyplugin"
        / "DummyPlugin-12345.local"
    )
    try:
        out_path.unlink()
    except FileNotFoundError:
        pass
    plugin_cli(["dummyplugin", "the_number=13"])
    interactive = capsys.readouterr().out
    assert dummy_plugin.plugin_output_file == out_path
    assert out_path.exists()
    with out_path.open() as f:
        assert interactive == f.read()
    out_path.unlink()
    plugin_cli(["dummyplugin", "the_number=13", "--batchmode"])
    output = capsys.readouterr().out
    assert "tail -f" in output
    out_f = Path([o.split()[-1] for o in output.split("\n") if "tail" in o][0])
    assert out_f.exists()
    with out_f.open() as f:
        assert "pending" in f.read()

    import re, freva

    pattern = r"'outputdir': '([^']*)'"

    plugin_cli(["dummypluginfolders", "variable=pr"])
    output_str = capsys.readouterr().out
    match = re.search(pattern, output_str)
    if match:
        folder_path = match.group(1)
    assert str(freva.history()[0]["id"]) in os.path.basename(
        os.path.normpath(folder_path)
    )

    plugin_cli(["dummypluginfolders", "variable=pr", "--unique-output", "true"])
    output_str = capsys.readouterr().out
    match = re.search(pattern, output_str)
    if match:
        folder_path = match.group(1)
    assert str(freva.history()[0]["id"]) in os.path.basename(
        os.path.normpath(folder_path)
    )

    plugin_cli(["dummypluginfolders", "variable=pr", "--unique-output", "false"])
    output_str = capsys.readouterr().out
    match = re.search(pattern, output_str)
    if match:
        folder_path = match.group(1)
    assert str(freva.history()[0]["id"]) not in os.path.basename(
        os.path.normpath(folder_path)
    )


def test_killed_jobs_set_to_broken():
    import freva

    proc = mp.Process(
        target=freva.run_plugin,
        args=("dummyplugin",),
        kwargs={"the_number": 10, "other": -10},
    )
    proc.start()
    time.sleep(1)
    proc.terminate()
    time.sleep(1)
    hist = freva.history()[0]
    assert hist["status_dict"][hist["status"]].lower() == "broken"


@mock.patch("os.getpid", lambda: 12345)
def test_list_tools(capsys, dummy_env):
    from freva.cli.plugin import main as run

    run(["--list-tools"])
    plugin_list = capsys.readouterr().out
    assert "DummyPlugin" in plugin_list
    run_cli(["plugin", "dummyplugin", "--doc"])
    help_str = capsys.readouterr().out
    assert similar_string(
        help_str,
        """DummyPlugin (v0.0.0): A dummy plugin
Options:
number                  (default: <undefined>)
                        This is just a number, not really important
the_number              (default: <undefined>) [mandatory]
                        This is *THE* number. Please provide it
something               (default: test)
                        No help available.
other                   (default: 1.4)
                        No help available.
input                   (default: <undefined>)
                        An input file
variable                (default: tas)
                        An input variable
extra_scheduler_options (default: --qos=test, --array=20)
                        Set additional options for the job submission to the
                        workload manager (, seperated). Note: batchmode and web
                        only.""",
    )


@mock.patch("os.getpid", lambda: 12345)
def test_run_pyclientplugin(dummy_history):
    import freva
    from evaluation_system.misc import config

    res, _ = freva.run_plugin("dummyplugin", the_number=32, caption="Some caption")
    assert res == 0
    _, res = freva.run_plugin("dummyplugin", the_number=32, show_config=True)
    res = "\n".join([l.strip() for l in res.split("\n") if l.strip()])
    assert similar_string(
        res,
        """    number: -the_number: 32 something: test other: 1.4 input: -variable: tas
extra_scheduler_options: - (default: )""",
    )
    return_val, repo = freva.run_plugin("dummyplugin", the_number=32, repo_version=True)
    assert "repository" in repo.lower()
    assert "version" in repo.lower()

    with pytest.raises(PluginNotFoundError):
        freva.run_plugin("dummyplugin0")


@mock.patch("os.getpid", lambda: 12345)
def test_run_plugin(capsys, dummy_history, dummy_env):
    from evaluation_system.misc import config

    with pytest.raises(ValidationError):
        run_cli(["plugin", "dummyplugin"])
    # test run tool
    run_cli(["plugin", "dummyplugin", "the_number=32", '--caption="Some caption"'])
    output_str = capsys.readouterr().out
    assert "Dummy tool was run" in output_str
    assert "the_number" in output_str
    # test get version
    run_cli(["plugin", "dummyplugin", "--repo-version"])
    # test batch mode
    run_cli(["plugin", "dummyplugin", "the_number=32", "--batchmode"])
    # test save config
    run_cli(["plugin", "dummyplugin", "the_number=32", "--save", "--debug"])
    fn = (
        Path(config.get(config.BASE_DIR_LOCATION))
        / "config/dummyplugin/dummyplugin.conf"
    )
    assert not fn.is_file()
    # test show config
    run_cli(["plugin", "dummyplugin", "the_number=42", "--show-config"])
    output_str = capsys.readouterr().out
    for line in (
        "number: -",
        "the_number: 42",
        "something: test",
        "other: 1.4",
        "input: -",
    ):
        assert line in output_str



from evaluation_system import __version__
from ._user_data import UserData
from ._databrowser import databrowser, count_values, facet_search
from ._plugin import (
    run_plugin,
    list_plugins,
    plugin_doc,
    read_plugin_cache,
    get_tools_list,
)
from ._esgf import esgf
from ._history import history
import warnings

warnings.filterwarnings("always", category=DeprecationWarning, module="freva.*")


__all__ = [
    "UserData",
    "databrowser",
    "count_values",
    "facet_search",
    "run_plugin",
    "list_plugins",
    "plugin_doc",
    "esgf",
    "history",
]



"""Tests for ingesting user data."""
from __future__ import annotations
from pathlib import Path
import datetime
import mock
from tempfile import TemporaryDirectory, NamedTemporaryFile
from typing import Generator, Tuple

import cftime
import pytest
import numpy as np
import xarray as xr


def create_data(
    num_files: int,
    variable_names: list[str],
    chunk_size: Tuple[int, ...],
    dims: Tuple[str, ...],
) -> Generator[Path, None, None]:
    """Create a netcdf dataset."""
    coords = {d: np.ones(chunk_size[n]) for (n, d) in enumerate(dims)}
    data = xr.Dataset()
    attrs = {}
    ntstep = 0
    with TemporaryDirectory() as temp_dir:
        for nfile in range(num_files):
            if "time" in dims:
                units = "seconds since 1970-01-01 00:00:00"
                coords["time"] = cftime.num2date(
                    np.array([ntstep + (j * 3600) for j in range(chunk_size[0])]),
                    units=units,
                )
                data["time"] = xr.DataArray(
                    data=coords["time"],
                    dims=("time",),
                    coords={"time": coords["time"]},
                    name="time",
                )
                attrs["frequency"] = "not valid"
            for variable_name in variable_names:
                dset = xr.DataArray(
                    np.zeros(chunk_size),
                    dims=dims,
                    coords=coords,
                    name=variable_name,
                )
                data[variable_name] = dset
                out_f = Path(temp_dir) / f"outfile_{nfile}.nc"
                data.set_coords(dims).to_netcdf(str(out_f))
            ntstep += chunk_size[0] * 3600
        yield Path(temp_dir)


@pytest.fixture(scope="function")
def valid_data_files() -> Generator[Path, None, None]:
    yield from create_data(
        10,
        ["tas", "time_bnds", "X", "rotated_pole", "longitude"],
        (10, 10, 10),
        ("time", "lat", "lon"),
    )


@pytest.fixture(scope="function")
def invalid_data_files() -> Generator[Path, None, None]:
    yield from create_data(1, ["tas", "foo"], (10, 10), ("lat", "lon"))


def test_invalid_data_files(invalid_data_files: Path, time_mock: mock_datetime) -> Path:
    from evaluation_system.api.user_data import DataReader

    in_file = list(invalid_data_files.rglob("*.*"))[0]
    data_reader = DataReader(invalid_data_files)
    with pytest.raises(ValueError):
        data_reader.get_metadata(in_file)
    not_a_nc_file = invalid_data_files / "not_a_nc_file.nc"
    not_a_nc_file.touch()
    data_reader = DataReader(not_a_nc_file)
    with pytest.raises(ValueError):
        data_reader.get_metadata(not_a_nc_file)


def test_add_valid_data(valid_data_files: Path, time_mock: mock_datetime) -> None:
    from evaluation_system.api.user_data import DataReader

    in_file = list(valid_data_files.rglob("*.*"))[0]
    data_reader = DataReader(valid_data_files)
    data = data_reader.get_metadata(in_file)
    assert data["variable"] == "tas"
    assert data["time_frequency"] == "hr"
    data_reader = DataReader(valid_data_files, variable="foo")
    data = data_reader.get_metadata(in_file)
    assert data["variable"] == "foo"
    assert data["time_frequency"] == "hr"
    with xr.open_dataset(str(in_file)) as dset:
        dset = dset.isel(time=0).load()
    dset.to_netcdf(in_file)
    data = data_reader.get_metadata(in_file)
    assert data["variable"] == "foo"
    assert data["time_frequency"] == "fx"


def test_get_time_frequency(valid_data_files: Path, time_mock: mock_datetime) -> None:
    from evaluation_system.api.user_data import DataReader

    data_reader = DataReader(Path("foo/bar.nc"))
    hour = 3600
    day = hour * 24
    mon = day * 30
    year = day * 360
    assert data_reader.get_time_frequency(0, "foo") == "fx"
    assert data_reader.get_time_frequency(0, "hr") == "hr"
    assert data_reader.get_time_frequency(60) == "subhr"
    assert data_reader.get_time_frequency(hour) == "hr"
    assert data_reader.get_time_frequency(day) == "day"
    assert data_reader.get_time_frequency(15 * day) == "sem"
    assert data_reader.get_time_frequency(mon) == "mon"
    assert data_reader.get_time_frequency(year) == "yr"


def test_get_file_name_from_metadata(
    valid_data_files: Path, time_mock: mock_datetime
) -> None:
    from evaluation_system.api.user_data import DataReader

    in_file = list(valid_data_files.rglob("*.*"))[0]
    defaults = dict(
        project="foo",
        product="fumanshu",
        institute="tong",
        model="mrfu",
        experiment="foo-boo",
        cmor_table="foo",
        ensemble="bar",
        realm="foo-kingdom",
        version="v0",
    )
    data_reader = DataReader(valid_data_files, **defaults)
    file_part = "foo/fumanshu/tong/mrfu/foo-boo/hr/foo-kingdom/foo/bar/v0/tas"
    assert file_part in str(data_reader.file_name_from_metdata(in_file).parent)
    defaults["variable"] = "foob"
    defaults["time_frequency"] = "3h"
    defaults.pop("cmor_table")
    data_reader = DataReader(valid_data_files, **defaults)
    file_part = "foo/fumanshu/tong/mrfu/foo-boo/3h/foo-kingdom/3h/bar/v0/foob"
    assert file_part in str(data_reader.file_name_from_metdata(in_file).parent)
    defaults.pop("version")
    data_reader = DataReader(valid_data_files, **defaults)
    file_part = "foo/fumanshu/tong/mrfu/foo-boo/3h/foo-kingdom/3h/bar/v19990909/foob"
    assert file_part in str(data_reader.file_name_from_metdata(in_file).parent)
    defaults.pop("ensemble")
    data_reader = DataReader(valid_data_files, **defaults)
    with pytest.raises(ValueError):
        data_reader.file_name_from_metdata(in_file)


def test_versions(valid_data_files: Path, time_mock: mock_datetime) -> None:
    def get_new_file(inp: Path, override: bool = False) -> Path:
        defaults = dict(
            project="foo",
            product="fumanshu",
            institute="tong",
            model="mrfu",
            experiment="foo-boo",
            cmor_table="foo",
            ensemble="bar",
            realm="foo-kingdom",
        )
        from evaluation_system.api.user_data import DataReader

        data_reader = DataReader(inp, **defaults)
        return data_reader.file_name_from_metdata(inp, override=override)

    in_file = list(valid_data_files.rglob("*.*"))[0]
    new_file = get_new_file(in_file)
    assert "v19990909" in str(new_file)
    new_file.parent.mkdir(exist_ok=True, parents=True)
    new_file.touch()
    time_mock._day = 10
    new_file2 = get_new_file(in_file)
    new_file2.parent.mkdir(exist_ok=True, parents=True)
    new_file2.touch()
    new_file3 = get_new_file(in_file, override=False)
    assert new_file != new_file2
    assert new_file3 == new_file2


def test_iter_data_files(valid_data_files: Path, time_mock: mock_datetime) -> None:
    from evaluation_system.api.user_data import DataReader

    input_files = list(valid_data_files.rglob("*.*"))
    not_a_nc_file = valid_data_files / "not_a_nc_file.txt"
    not_a_nc_file.touch()
    data_reader = DataReader(valid_data_files)
    files = [f for f in data_reader]
    data_reader = DataReader(valid_data_files / "*.nc")
    files = [f for f in data_reader]
    assert len(files) == len(input_files)
    data_reader = DataReader(valid_data_files / "*.txt")
    files = [f for f in data_reader]
    assert files == []
    data_reader = DataReader(not_a_nc_file)
    files = [f for f in data_reader]
    assert files == []


def test_link_my_data(dummy_crawl, dummy_plugin, valid_data_files, time_mock):
    import freva

    input_files = list(valid_data_files.rglob("*.nc"))
    dummy_plugin.add_output_to_databrowser(
        valid_data_files, "muh", "mah", experiment="foo"
    )
    assert len(list(freva.databrowser(experiment="foo"))) == len(input_files)
    assert len(list(freva.databrowser(product="muh.mah"))) == len(input_files)


def test_add_my_data(valid_data_files, time_mock):
    from freva import UserData, count_values
    from freva.cli.user_data import main as run

    defaults = [
        "--institute",
        "tong",
        "--model",
        "mrfu",
        "--experiment",
        "foo-boo",
        "--cmor_table",
        "foo",
        "--ensemble",
        "bar",
        "--realm",
        "foo-kingdom",
    ]
    run(["add", "foo-product", str(valid_data_files)] + defaults)
    run(["add", "foo-product", str(valid_data_files)] + defaults + ["--override"])
    input_files = list(valid_data_files.rglob("*.nc"))
    with pytest.raises(SystemExit):
        run(["add", "foo-product", str(valid_data_files)])
    with pytest.raises(ValueError):
        run(["add", "foo-product", str(valid_data_files), "-d"])

    user_data = UserData()
    assert count_values(product="foo-product") == len(input_files)
    with pytest.raises(ValueError):
        user_data.add("foo-product", valid_data_files, how="foo")
    with pytest.warns(UserWarning):
        user_data.add("foo-product")


def test_add_methods():
    from freva import UserData
    import shutil
    import os

    assert UserData._set_add_method("cp") == shutil.copy
    assert UserData._set_add_method("mv") == shutil.move
    assert UserData._set_add_method("ln") == os.symlink
    assert UserData._set_add_method("link") == os.link
    with pytest.raises(ValueError):
        UserData._set_add_method("foo")


def test_delete_my_data(valid_data_files, time_mock):
    import freva
    from freva import UserData
    from freva.cli.user_data import main as run
    from evaluation_system.model.user import User
    from evaluation_system.api.user_data import get_output_directory

    user_data = UserData()
    root_path = get_output_directory() / f"user-{User().getName()}" / "foo-product"
    assert freva.databrowser(product="foo-product", count=True) > 0
    nfiles = len(list(root_path.rglob("*.nc")))
    assert nfiles > 0
    run(["delete", str(root_path)])
    assert freva.databrowser(product="foo-product", count=True) == 0
    assert len(list(root_path.rglob("*.nc"))) > 0
    user_data.delete(root_path, delete_from_fs=True)
    assert len(list(root_path.rglob("*.nc"))) == 0


def test_index_my_data(dummy_crawl, capsys, dummy_env, valid_data_files, time_mock):
    from freva import UserData
    from evaluation_system.tests import run_cli
    from freva.cli.user_data import main as run
    from evaluation_system.misc.exceptions import ValidationError
    from evaluation_system.model.solr import SolrFindFiles

    run(["index", "--data-type=fs", "-d"])
    captured = capsys.readouterr()
    assert "Status: crawling ..." in captured.out
    assert "ok" in captured.out
    assert len(list(SolrFindFiles.search(product="foo"))) == len(dummy_crawl) - 2
    assert len(list(SolrFindFiles.search(product="foo", latest_version=False))) == len(
        dummy_crawl
    )
    user_data = UserData()
    with pytest.raises(NotImplementedError):
        user_data.index(dtype="something")
    with pytest.raises(SystemExit):
        with pytest.raises(ValidationError):
            run_cli(["user-data", "index", "/tmp/forbidden/folder"])

    import os, shutil

    tmp_files = [
        "observations.station/DWD/DWD/essen/1hr/cmorized/1hr/r1i1p1/v2281/pr/pr_1hr_DWD_essen_r1i1p1_189509010000-192412312300.nc",
        "observations.station/DWD/DWD/essen/1hr/cmorized/1hr/r1i1p1/v2281/pr/pr_1hr_DWD_essen_r1i1p1_192509010000-194112312300_as.nc",
        "observations.station/DWD/DWD/essen/1hr/cmorized/1hr/r1i1p1/v2281/pr/pr_1hr_DWD_essen_r1i1p1_199509010000-202112312300.nc",
    ]
    new_files = [str(user_data.user_dir) + "/" + file for file in tmp_files]
    for file in new_files:
        os.makedirs(os.path.dirname(file), exist_ok=True)
        with open(file, "w"):
            pass
    for abort_flag in ["--continue-on-errors", "--continue", "-c"]:
        run(["index", abort_flag])
        captured = capsys.readouterr()
        assert "Status: crawling ..." in captured.out
        assert "ok" in captured.out
        assert (
            len(
                list(
                    SolrFindFiles.search(
                        project="user*", product="observations.station"
                    )
                )
            )
            == len(tmp_files) - 1
        )
        run(["delete", str(user_data.user_dir) + "/observations.station/"])
    shutil.rmtree(str(user_data.user_dir) + "/observations.station/")


def test_wrong_datatype(dummy_crawl, capsys, dummy_env, time_mock):
    from evaluation_system.tests import run_cli
    from evaluation_system.model.solr import SolrFindFiles

    dummy_crawl.append(dummy_crawl[0].parent / "more_info" / dummy_crawl[0].name)
    dummy_crawl[-1].parent.mkdir(exist_ok=True, parents=True)
    dummy_crawl[-1].touch()
    with pytest.raises(ValueError):
        run_cli(["user-data", "index", "-d"])
    with pytest.raises(SystemExit):
        run_cli(["user-data", "index"])
    captured = capsys.readouterr()
    assert "ValueError" in captured.err
    assert len(list(SolrFindFiles.search())) == 0
    assert len(list(SolrFindFiles.search(latest_version=False))) == 0


def test_validate_path(root_path_with_empty_config, time_mock):
    from freva import UserData

    user_data = UserData()
    root_path_str = str(root_path_with_empty_config)
    print(root_path_str)
    assert user_data._validate_user_dirs(root_path_str) == (
        root_path_with_empty_config,
    )
